#!/usr/bin/env python
# coding: utf-8

from __future__ import print_function, unicode_literals

import re
import os
import sys
import time
import json
import zlib
import random
import logging
import argparse
import colorsys
from datetime import datetime
from PIL import ImageFont, ImageDraw, Image

"""
softchat.py: convert twitch/youtube chat into softsubs
2020-10-18, v0.6, ed <irc.rizon.net>, MIT-Licensed
https://ocv.me/dev/?softchat.py

==[ NOTES / README ]===================================================

 - superchats will display for 2x the time and with inverted colors

 - moderator messages are emphasized
     (larger outline, and prefixed with a ball)

 - mode 1, sidebar chat, creates a huge amount of subtitle events
     which many media players (including mpv) will struggle with

     for example, seeking will take like 5sec

     you can fix this by muxing the subtitle into the vid:
     ffmpeg -i the.webm -i the.ass -c copy muxed.mkv

 - mode 2, danmaku, will look blurry and/or jerky in most players
     unless you have the subtitles render at your native screen res

     for example in mpv you could add these arguments:
     --vo=direct3d --sub-delay=-3 --vf=fps=90

     replace 90 with your monitor's fps

==[ DEPENDENCIES ]=====================================================
each version below is the latest as of writing,
tested on cpython 3.8.1

 - chat rips made using the modified chat_replay_downloader.py
     mod: https://ocv.me/dev/?chat_replay_downloader-3.py
     orig: https://github.com/xenova/chat-replay-downloader/

 - all the noto fonts in a subfolder called noto-hinted
     see https://www.google.com/get/noto/
     or just https://noto-website-2.storage.googleapis.com/pkgs/Noto-hinted.zip
     (remove "un" from "unhinted" in the download link)

 - REQUIRED:
   python -m pip install --user pillow

 - OPTIONAL (not yet implemented here):
   python -m pip install --user git+https://github.com/googlefonts/nototools.git@v0.2.13#egg=nototools

 - OPTIONAL (recommended):
   python -m pip install --user "fugashi[unidic]"
   python -m unidic download

 - OPTIONAL (this also works with some modifications):
   python -m pip install --user mecab-python3
   # bring your own dictionaries

==[ HOW-TO / EXAMPLE ]=================================================

 1 download the youtube video, for example with youtube-dl or
   https://ocv.me/dev/?ytdl-tui.py

   it will eventually say "Merging formats into some.mkv",
   use that filename below except replace extension as necessary
 
 2 download the chatlog:
   chat_replay_downloader.py https://youtu.be/fgsfds -output "some.json"

 3 convert the chatlog into subtitles (-m2=danmaku)
   softchat.py -m2 "some.json"

 4 play the video (with --vf=fps=FRAMERATE due to danmaku)
   mpv some.mkv --sub-files="some.json.ass" --sub-delay=-3 --vf=fps=60

==[ NEW ]==============================================================

 - convert to hiragana instead of katakana
 - fix trailing blank lines in sidebar mode
 - slightly better fix for the mecab dll location
 - examples/notes at the bottom

==[ TODO ]=============================================================

 - build optimal font using noto-merge-fonts
 - per-line background shading (optional)
 - more stuff probably

==[ RELATED ]==========================================================

 - https://ocv.me/dev/?gist/chat-heatmap.py

"""


try:
    from nototools import merge_fonts

    HAVE_NOTO_MERGE = True
except ImportError:
    HAVE_NOTO_MERGE = False


debug = logging.debug
info = logging.info
warn = logging.warning
error = logging.error


LINUX = sys.platform.startswith("linux")
WINDOWS = sys.platform == "win32"
MACOS = sys.platform == "darwin"


class LoggerFmt(logging.Formatter):
    def format(self, record):
        if record.levelno == logging.DEBUG:
            ansi = "\033[01;30m"
        elif record.levelno == logging.INFO:
            ansi = "\033[0;32m"
        elif record.levelno == logging.WARN:
            ansi = "\033[0;33m"
        else:
            ansi = "\033[01;31m"

        ts = datetime.utcfromtimestamp(record.created)
        ts = ts.strftime("%H:%M:%S.%f")[:-3]
        return f"\033[0;36m{ts}{ansi} {record.msg}\033[0m"


if __name__ == '__main__':
    if WINDOWS:
        os.system("")

    logging.basicConfig(
        level=logging.INFO,  # INFO DEBUG
        format="\033[36m%(asctime)s.%(msecs)03d\033[0m %(message)s",
        datefmt="%H%M%S",
    )
    lh = logging.StreamHandler(sys.stderr)
    lh.setFormatter(LoggerFmt())
    logging.root.handlers = [lh]


try:
    # help python find libmecab.dll, adjust this to fit your env,
    # TODO make this fix less bad (linux/macos, python versions, sys/user)
    home = os.path.expanduser('~')
    dll_path = os.path.join(home, r'AppData\Roaming\Python\lib\site-packages\fugashi')

    os.add_dll_directory(dll_path)
    from fugashi import Tagger
    
    dicrc = os.path.join(dll_path, 'dicrc')
    with open(dicrc, 'wb') as f:
        f.write('\n'.join([
            r'node-format-yomi = %f[9] ',
            r'unk-format-yomi = %m',
            r'eos-format-yomi  = \n',
            ''
        ]).encode('utf-8'))

    wakati = Tagger('-Owakati')
    yomi = Tagger('-Oyomi -r ' + dicrc.replace('\\', '\\\\'))

    #import MeCab
    #wakati = MeCab.Tagger('-Owakati')
    HAVE_TOKENIZER = True
    info('found mecab')
except:
    HAVE_TOKENIZER = False
    import traceback

    warn('could not load mecab:\n' + traceback.format_exc() + '-' * 72 + '\n')


class TextStuff(object):
    def __init__(self, sz):
        self.sz = sz
        self.font = ImageFont.truetype("NotoSansCJKjp-Regular.otf", size=sz)
        self.im = Image.new("RGB", (3840, 2160), "white")
        self.imd = ImageDraw.Draw(self.im)
        self.pipe_width = self.font.getsize('|')[0]
        self.font_ofs = self.font.getmetrics()[1]
        self.cache = {}
        self.vsize = self.caching_vsize

    def vsize_impl(self, text):
        w, h = self.imd.textsize('|' + text.replace('\n', '\n|'), self.font)
        return w - self.pipe_width, h
    
    def caching_vsize(self, text):
        if len(text) > 24:
            return self.vsize_impl(text)
        
        # faster than try/catch and get(text,None)
        if text in self.cache:
            return self.cache[text]
        
        ret = self.vsize_impl(text)
        self.cache[text] = ret
        return ret

    def unrag(self, text, width):
        """
        copied from  http://xxyxyz.org/line-breaking/
        license clarified per email:
          BSD or MIT since the EU doesn't recognize Public Domain
        
        only change so far is replacing len(w) with vsize(w+"_")[0]
        """
        words = text.split()
        count = len(words)
        offsets = [0]
        for w in words:
            offsets.append(offsets[-1] + self.vsize(w + "_")[0])

        minima = [0] + [10 ** 20] * count
        breaks = [0] * (count + 1)
        for i in range(count):
            j = i + 1
            while j <= count:
                w = offsets[j] - offsets[i] + j - i - 1
                if w > width:
                    break
                cost = minima[i] + (width - w) ** 2
                if cost < minima[j]:
                    minima[j] = cost
                    breaks[j] = i
                j += 1

        lines = []
        j = count
        while j > 0:
            i = breaks[j]
            lines.append(u" ".join(words[i:j]))
            j = i
        lines.reverse()
        return lines


def hms(s):
    h, s = divmod(s, 3600)
    m, s = divmod(s, 60)
    return '{:d}:{:02d}:{:05.2f}'.format(int(h), int(m), s)


def assan(x):
    return x.replace("{", "<").replace("}", ">").replace('\\', '\\{}')


def main():
    t0_main = time.time()

    random.seed(b'nope')

    ap = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="convert modified chat_replay_downloader.py json into box-confined or danmaku-style softsubs")

    ap.add_argument("-m", metavar='MODE', type=int, default=1, help="mode, 1=box, 2=danmaku")
    ap.add_argument("-r", metavar='WxH', type=str, default="1280x720", help="video res")
    ap.add_argument("-b", metavar='WxH+X+Y', type=str, default=None, help="subtitle area")
    ap.add_argument("-f", action='store_true', help='fill chat background')
    ap.add_argument("--sz", metavar='POINTS', type=int, default=0, help="font size")
    ap.add_argument("--spd", metavar='SPEED', type=int, default=256, help="[danmaku] pixels/sec")
    ap.add_argument("--kana", action="store_true", help="convert kanji to kana")
    ap.add_argument("fn", metavar='JSON_FILE')
    ar = ap.parse_args()

    if ar.kana and not HAVE_TOKENIZER:
        error('you requested --kana but mecab failed to load')
        sys.exit(1)

    if not ar.sz:
        ar.sz = 24 if ar.m == 1 else 32
        info(f"fontsize {ar.sz} pt")

    vw, vh = [int(x) for x in ar.r.split('x')]

    bw, bh, bx, by = [int(x) for x in re.split(r'[x,+]+', ar.b if ar.b else ar.r + '+0+0')]

    z = TextStuff(ar.sz)
    fofs = z.font_ofs

    info(f"loading {ar.fn}")
    with open(ar.fn, 'r', encoding='utf-8') as f:
        jd = json.load(f)

    ptn_kanji = re.compile(r'[\u4E00-\u9FAF]')
    ptn_kana = re.compile(r'[\u3040-\u30FF]')
    ptn_ascii = re.compile(r'[a-zA-Z]')
    ptn_pre = re.compile(r'([　、。〇〉》」』】〕〗〙〛〜〞〟・…⋯！＂）＊－．／＞？＠＼］＿～｡｣･￭￮]+)')
    ptn_post = re.compile(r'([〈《「『【〔〖〘〚〝（＜［｀｢]+)')

    msgs = []
    info(f"converting {len(jd)} chat entries")
    last_msg = None
    for n_msg, msg in enumerate(jd):
        txt = msg['message'] or '--'
        cmp_msg = f"{msg['author_id']}\n{txt}"
        if last_msg == cmp_msg:
            continue

        last_msg = cmp_msg

        n_ascii = len(ptn_ascii.findall(txt))
        n_kanji = len(ptn_kanji.findall(txt))
        n_kana = len(ptn_kana.findall(txt))
        
        # if the amount of ascii compared to kanji/kana
        # is less than 30%, assume we'll need MeCab
        is_jp = (n_kanji + n_kana) / (n_kanji + n_kana + n_ascii + 0.1) > 0.7

        # transcription from kanji to kana if requested
        if ar.kana and is_jp and n_kanji:
            txt2 = yomi.parse(txt)
            txt = ''
            for ch in txt2:
                co = ord(ch)
                # hiragana >= 0x3041 and <= 0x3096
                # katakana >= 0x30a1 and <= 0x30f6
                if co >= 0x30a1 and co <= 0x30f6:
                    txt += chr(co - (0x30a1 - 0x3041))
                else:
                    txt += ch

        if ar.m == 1:
            wrap_width = bw
        else:
            # vsfilter wraps it anyways orz
            wrap_width = vw / 2

        #if n_msg % 100 == 0 and len(z.cache) > 1024 * 1024:
        #    z.cache = {}
        
        # wrap to specified width
        # by splitting on ascii whitespace
        vtxt = z.unrag(txt, bw)
        vsz = z.vsize('\n'.join(vtxt))
        
        if vsz[0] >= bw and is_jp:
            # too wide, is japanese,
            # try wrapping on japanese punctuation instead
            vtxt = ptn_pre.sub("\\1\n", txt)
            vtxt = ptn_post.sub("\n\\1", vtxt)
            vtxt = vtxt.split('\n')
            vsz = z.vsize('\n'.join(vtxt))

            if vsz[0] >= bw and HAVE_TOKENIZER:
                # still too wide, wrap on word-boundaries
                vtxt = z.unrag(wakati.parse(txt), bw)
                vtxt = [x.replace(' ', '') for x in vtxt]

                for n in range(1, len(vtxt)):
                    # move most punctuation to the prev line
                    ln = vtxt[n]
                    m = ptn_pre.search(ln)
                    if m and m.start() == 0:
                        vtxt[n-1] += ln[:m.end()]
                        vtxt[n] = ln[m.end():]

                vsz = z.vsize('\n'.join(vtxt))

        vtxt = [x for x in vtxt if x.strip()]

        if n_msg % 100 == 0:
            info('  {} / {}   {}%   {}   {}\n   {}\n'.format(
                n_msg,
                len(jd),
                int((n_msg*100)/len(jd)),
                msg["time_text"],
                vsz,
                '\n   '.join(vtxt)))

        o = {
            "name": msg["author"],
            "t0": msg["time_in_seconds"],
            "sx": vsz[0],
            "sy": vsz[1],
            "txt": vtxt
        }

        if 'amount' in msg:
            o['shrimp'] = msg['amount']
            o['color'] = msg['body_color']['hex'][1:][:-2]  # "#1de9b6ff"

        if 'badges' in msg:
            o['badges'] = msg['badges']
        
        msgs.append(o)

        # 40000 01:54:37 SilkTouch is an enchant for picaxe
        #if n_msg > 40000:
        #    break
    
    if ar.f:
        opts = {
            'back': '80',
            'shad': 'ff',
            'opaq': '3'
        }
    else:
        opts = {
            'back': '00',
            'shad': '80',
            'opaq': '1'
        }

    vis = []
    out_fn = ar.fn + ".ass"
    info(f"creating {out_fn}")
    with open(out_fn, "wb") as f:
        f.write("""\
[Script Info]
Title: softchat.py
ScriptType: v4.00+
WrapStyle: 0
ScaledBorderAndShadow: yes
YCbCr Matrix: None
PlayResX: {vw}
PlayResY: {vh}

[Aegisub Project Garbage]
Last Style Storage: Default
Video File: ?dummy:30.000000:40000:{vw}:{vh}:47:163:254:
Video AR Value: 1.777778
Video Zoom Percent: 1.000000

[V4+ Styles]
Format: Name, Fontname, Fontsize, PrimaryColour, SecondaryColour, OutlineColour, BackColour, Bold, Italic, Underline, StrikeOut, ScaleX, ScaleY, Spacing, Angle, BorderStyle, Outline, Shadow, Alignment, MarginL, MarginR, MarginV, Encoding
Style: a,Noto Sans CJK JP Regular,{sz},&H00FFFFFF,&H000000FF,&H{back}000000,&H{shad}000000,0,0,0,0,100,100,0,0,{opaq},2,1,7,0,0,0,1

[Events]
Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
""".format(vw=vw, vh=vh, sz=ar.sz, **opts).encode('utf-8'))

        # Dialogue: 0,0:00:00.00,0:00:05.00,a,,0,0,0,,hello world

        n_msg = 0
        msg = None
        supers = []
        colormap = {}
        for next_msg in msgs + [None]:
            if not msg or (next_msg and next_msg["t0"] <= 0):
                msg = next_msg
                continue

            n_msg += 1
            if n_msg % 1000 == 1:
                info(f"writing {hms(msg['t0'])}, #{n_msg} / {len(msgs)}")
            
            name = msg["name"]
            c = colormap.get(name, None)
            if not c:
                if ar.m == 1:
                    bri = 0.5
                    sat = 1
                else:
                    bri = 0.4
                    sat = 0.8

                c = zlib.crc32(name.encode('utf-8')) & 0xffffffff
                r, g, b = [int(x*255) for x in colorsys.hsv_to_rgb((c % 256) / 256.0, sat, bri)]
                c = f'{r:02x}{g:02x}{b:02x}'
                colormap[name] = c

            shrimp = None
            if 'shrimp' in msg:
                c2 = msg['color']
                c2 = f"{c2[4:6]}{c2[2:4]}{c2[0:2]}"  # thx ass
                shrimp = rf"{{\bord4\shad4\3c&H{c2}&\c&H000000&}}{msg['shrimp']}"

            name = assan(name)
            msg["txt"] = [assan(x) for x in msg["txt"]]

            if ar.m == 1:
                # text = colored name followed by the actual lines, ass-escaped
                txt = [rf"{{\3c&H{c}&}}{name}"]
                if shrimp:
                    txt.append(shrimp)

                txt.extend(msg["txt"])
                
                # show new messages bottom-left (heh)
                msg["txt"] = txt
                msg["px"] = bx
                msg["py"] = by + bh - msg["sy"] - fofs / 2  # or something idgi
                
                ta = hms(msg["t0"])
                tb = hms(next_msg["t0"] if next_msg else msg["t0"] + 10)

                rm = 0
                for m in vis:
                    m["py"] -= msg["sy"]
                    if m["py"] + m["sy"] < by:
                        #debug('drop {} at {}'.format(hms(m["t0"]), ta))
                        rm += 1

                vis = vis[rm:] + [msg]

                for m in vis:
                    step = m["sy"] / len(m["txt"])
                    x = m["px"]
                    y = m["py"]
                    for ln in m["txt"]:
                        txt = r'{{\pos({},{})}}{}'.format(x, y, ln)
                        x = m["px"] + 8
                        y += step
                        f.write("Dialogue: 0,{},{},a,,0,0,0,,{}\n".format(
                            ta, tb, txt).encode('utf-8'))
            else:
                txts, t0, w, h = [msg[x] for x in ["txt", "t0", "sx", "sy"]]
                txt = '\\N'.join(txts)

                #if txt.startswith('SilkTouch is an enchant for picaxe'):
                #    print('a')
                
                # pillow height calculation is off by a bit; this is roughly it i think
                h = int(h + fofs * 0.5)

                # ass linespacing is huge, compensate (wild guess btw)
                h += ar.sz * 0.25 * (len(txts) - 1)
                
                shrimp_mul = 1
                if shrimp:
                    txt = f"{shrimp} {txt}"
                    shrimp_mul = 2
                    w += h * 4  # donation is not included, TODO maybe

                if 'badges' in msg and 'Moderator' in msg['badges']:
                    # mods and other VIPs
                    txt = rf"{{\bord24\shad6}}*{{\bord6}}{txt}"

                len_boost = 0.5
                td = (vw + w - w * len_boost) * shrimp_mul / ar.spd
                t1 = t0 + td

                # collision detection polls at 50% and 80% from the left,
                # figure out timestamps when the right-side hits those
                abs_spd = (vw + w * 2.0) / td
                p8 = t0 + (w + vw * 0.2) / abs_spd
                p5 = t0 + (w + vw * 0.5) / abs_spd

                # and when the left-side hits those,
                # to compare against other p8/p5's and find a free slot
                a8 = t0 + (vw * 0.2) / abs_spd
                a5 = t0 + (vw * 0.5) / abs_spd

                rm = []
                taken8 = []
                taken5 = []
                for m in vis:
                    m_py, m_sy, m_p8, m_p5 = [m[x] for x in ["py", "sy", "p8", "p5"]]

                    if t0 > m_p5:
                        rm.append(m)
                        continue

                    if a8 < m_p8:
                        taken8.append([m_py, m_py + m_sy, m_p8, m])

                    if a5 < m_p5:
                        taken5.append([m_py, m_py + m_sy, m_p5, m])

                for m in rm:
                    vis.remove(m)

                ymax = vh - h
                frees = [[ymax, 0, ymax]]  # size, y0, y1
                for y1, y2, _, _ in sorted(taken8):
                    rm = []
                    add = []
                    for free in frees:
                        _, fy1, fy2 = free

                        if fy1 >= y2 or y1 >= fy2:
                            # does not intersect
                            continue

                        elif fy1 <= y1 and fy2 >= y2:
                            # sub slices range in half, split it
                            rm.append(free)
                            
                            nfsz = y1 - fy1
                            if nfsz > h / 3.0:
                                # h/3 to ensure less than 33% overlap
                                add.append([nfsz, fy1, y1])

                            nfsz = fy2 - y2
                            if nfsz > h / 3.0:
                                add.append([nfsz, y2, fy2])

                        elif y2 >= fy1:
                            # sub slices top of free
                            rm.append(free)
                            
                            nfsz = fy2 - y2
                            if nfsz > h / 3.0:
                                add.append([nfsz, y2, fy2])

                        elif y1 <= fy2:
                            # slices bottom
                            rm.append(free)

                            nfsz = y1 - fy1
                            if nfsz > h / 3.0:
                                add.append([nfsz, fy1, y1])

                        else:
                            raise Exception("fug")
                
                    for x in rm:
                        frees.remove(x)
                    
                    frees.extend(add)

                if not frees:
                    # can't be helped, pick a random y to collide in
                    y = int(random.randrange(ymax))
                    best = [2, y-1, y+1]
                else:
                    avail, y0, y1 = sorted(frees)[-1]
                    if avail <= h:
                        y = int(y0 + avail / 2)
                    else:
                        # just centering looks boring, let's rand
                        y = y0 + random.randrange(avail - h)

                msg["t1"] = t1
                msg["p8"] = p8
                msg["p5"] = p5
                msg["px"] = vw
                msg["py"] = y
                msg["txt"] = rf"{{\move({vw},{y},{-w},{y})\3c&H{c}&}}{txt}{{\fscx40\fscy40\bord1}}\N{name}"
                vis.append(msg)

                ln = "Dialogue: 0,{},{},a,,0,0,0,,{}\n".format(
                    hms(t0), hms(msg["t1"]), msg["txt"]).encode('utf-8')

                if shrimp:
                    supers.append(ln)
                else:
                    f.write(ln)

            msg = next_msg

        for ln in supers:
            f.write(ln)

    #from pprint import pprint; pprint(msgs[-5:])
    t1_main = time.time()
    info(f'finished in {t1_main-t0_main:.2f} sec')


if __name__ == '__main__':
    main()


r"""

===[ side-by-side with kanji transcription ]===========================

chat_replay_downloader.py https://www.youtube.com/watch?v=7LXgVlrfsWw -output ars-37-minecraft-3.json

softchat.py ..\yt\ars-37-minecraft-3.json -b 340x500+32+32 && copy /y ..\yt\ars-37-minecraft-3.json.ass ..\yt\ars-37-minecraft-3.json.1.ass

softchat.py ..\yt\ars-37-minecraft-3.json -b 340x500+360+32 --kana && copy /y ..\yt\ars-37-minecraft-3.json.ass ..\yt\ars-37-minecraft-3.json.2.ass

(head -n 21 ars-37-minecraft-3.json.1.ass ; (tail -n +22 ars-37-minecraft-3.json.1.ass; tail -n +22 ars-37-minecraft-3.json.2.ass) | sort) > ars-37-minecraft-3.json.ass

c:\users\ed\bin\mpv.com "..\yt\ars-37-minecraft-3.json.mkv" -ss 1:15:20


===[ re-parse a downloaded youtube json ]==============================

chat_replay_downloader.py https://www.youtube.com/watch?v=0Qygvs0rG50 -output ame-minecraft-railway-research.json
chat_replay_downloader.py ame-minecraft-railway-research.json.json -output ame-minecraft-railway-research-2.json


===[ danmaku / nnd-style ]=============================================

# assumes your mpv config is interpolation=no, blend-subtitles=no
..\dev\softchat.py -m 2 "ame-minecraft-railway-research-2.json" && C:\Users\ed\bin\mpv.com ame-minecraft-railway-research-2.json.mkv --vo=direct3d --vf=fps=90 --sub-delay=-3

# sanic
--vo=direct3d --sub-delay=-3 --vf=fps=75 --speed=1.2

# alternate display mode, usually worse
C:\Users\ed\bin\mpv.com ame-minecraft-railway-research-2.json.mkv --interpolation=yes --blend-subtitles=yes --video-sync=display-resample --tscale=mitchell --hwdec=off --vo=direct3d

# use --sub-files=some.ass to specify a sub with another name


===[ junk ]============================================================

..\dev\softchat.py -m 1 -b 320x600+960+32 ame-minecraft-railway-research-2-2.json && C:\Users\ed\bin\mpv.com ame-minecraft-railway-research-2.json.mkv --vo=direct3d --sub-files=ame-minecraft-railway-research-2-2.json.ass --sub-delay=-3 -ss 120

cpy3: 15.46 sec @15k NOcache
cpy3:  8.87 sec @15k 16k 24

pypy: 18.75 sec @15k NOcache
pypy: 15.32 sec @15k 128 32
pypy: 13.32 sec @15k  1k 32
pypy: 12.59 sec @15k  4k 64
pypy: 12.23 sec @15k  4k 16
pypy: 12.09 sec @15k  4k 32
pypy: 12.10 sec @15k  8k 24
pypy: 11.42 sec @15k 16k 64
pypy: 11.39 sec @15k 16k 16
pypy: 11.26 sec @15k 16k 32
pypy: 11.23 sec @15k 16k 24

cpy3 full: 46.45 unb 24
cpy3 full: 48.56 32k 24
pypy full: 54.92 unb 24
pypy full: 57.34 32k 24

grep -E '"(amount|hex|message|time_text)":|^    \},' ame-minecraft-railway-research-2.json | grep -E '"amount":' -C5 | less

"""
